import 'package:djinn_flutter/src/djinn_service.dart';

class Djinn {
  Map<String, ServiceRegister> _services = {};
  Map<String, DjinnService> _singletons = {};

  static final Djinn _instance = Djinn._internal();
  factory Djinn() => _instance;

  Djinn._internal();

  void register<T extends DjinnService>(DjinnServiceRegister<T> serviceFn,
      [bool scoped = false]) {
    final serviceName = T.toString();

    if (_services[serviceName] != null) {
      print('Warning: Service $serviceName is already registered.');
      return;
    }

    _services[serviceName] = ServiceRegister(serviceFn, scoped);
  }

  T get<T extends DjinnService>() {
    final serviceName = T.toString();

    final service = _services[serviceName];
    if (service == null) {
      print('Warning: Service $serviceName is not registered.');
      return null;
    }

    if (service.scoped) {
      final instance = service.serviceFn();
      instance.init();
      return instance;
    }

    return _singletons[serviceName];
  }

  void remove<T extends DjinnService>() {
    final serviceName = T.toString();
    _services.remove(serviceName);
    _singletons.remove(serviceName);
  }

  void init() {
    _services.entries
        .where((MapEntry<String, ServiceRegister> it) => !it.value.scoped)
        .forEach((MapEntry<String, ServiceRegister> it) {
      _singletons[it.key] = it.value.serviceFn();
    });

    _singletons.values.forEach((DjinnService it) => it.init());
  }
}

typedef T DjinnServiceRegister<T extends DjinnService>();

class ServiceRegister {
  DjinnServiceRegister serviceFn;
  bool scoped;

  ServiceRegister(this.serviceFn, this.scoped);
}
